/*
	Simple IRC Client
	by Pat Niemeyer (pat@icon-stl.net)

	Attributes are :
		rows, cols, 	- The size of the box
		backlines  		- number of lines of stored history for scroll back
		server, port 	- IRC server and port number
		channel			- IRC channel name
		nick 		- If you don't specify nick a random one is generated.
		bigfont 	- Set font to 24 point
		keepalive	- Respond to PINGs from server.  This is turned off
					whenever you leave the applet, causing the connection
					to be dropped after a few minutes of inactivity.

		The IRC RFC can be found at:

			http://www.cis.ohio-state.edu/htbin/rfc/rfc1459.html
*/

package net.irc;

import awt.*;
import browser.*;
import java.io.*;
import net.*;
import text.*;

public class IRC extends Applet implements Runnable, TextCallBack {
	Thread runner;
	NetworkClient nc;
	PrintStream out;
	InputStream in;
	TextBox tb;
	TextEntry te;
	String server, channel, nick;
	int port;
	boolean bigfont = false, 
		keepalive = false;
	int backlines;
	private boolean STOPPED, DEBUG=true;
	int pad = 4;

	public void init() {
		int rows, cols;
		String s;

		s = getAttribute("bigfont");
		if (s != null) bigfont=true;
		s = getAttribute("keepalive");
		if (s != null) keepalive=true;
		s = getAttribute("channel");
		if (s != null) channel = s; else channel = "#irchelp";
		s = getAttribute("nick");
		if (s != null) nick = s; else nick = randnick();
		s = getAttribute("server");
		if (s != null) server = s; else server = "irc.colorado.edu";
		s = getAttribute("port");
		if (s != null) port = Integer.parseInt(s); else port = 6667;
		s = getAttribute("backlines");
		if (s != null) backlines = Integer.parseInt(s);
		s = getAttribute("rows");
		if (s != null) rows = Integer.parseInt(s); else rows = 20;
		s = getAttribute("cols");
		if (s != null) cols = Integer.parseInt(s); else cols = 80;

		System.out.println("Server: "+server+", port: "+port);
		System.out.println("rows: "+rows+", cols: "+cols);

		Font font;
		if (bigfont)
			font = getFont("Courier", Font.BOLD, 18);
		else
			font = getFont("Courier", Font.BOLD, 12);

		tb = new TextBox(this, font, rows, cols, backlines, 0, 0);
		te = new TextEntry(this, font, 80, 0, tb.Height + pad);
		resize( Math.max(tb.Width, te.Width), tb.Height + te.Height + pad );
	}

	public void start() {
		STOPPED=false;
		if ( runner == null ) {
			runner = new Thread(this);
			runner.start();
		}
	}

	private boolean connect() {
		System.out.println("Connecting...\n");
		tb.wait_notice("Connecting to "+server+" port "+port);
		
		try {
			nc = new NetworkClient(server, port);
		} catch (Exception e) {
			System.out.println("NetworkClient exception: "+e);
		}

		if ( nc == null ) {
			tb.notice("Error connecting to "+server);
			return false;
		}
	
		out = nc.serverOutput;
		in = nc.serverInput;

		System.out.println("Sending registration...\n");
		out.println("user "+nick+" machine.net irc.server :Foo Bat");
		System.out.println("user "+nick+" machine.net irc.server :Foo Bat");
		out.println("nick "+nick);
		System.out.println("nick "+nick);
		out.println("join "+channel);
		System.out.println("join "+channel);
		return true;
	}

	/* 
		Wait on a text entry of a nickname
	*/
	private synchronized void waitfornick() {
		tb.notice("Enter your nickname below");
		te.Prompt("Nickname: ");
		while ( nick.equals("PROMPT") )
			wait();
		te.Prompt();
	}

	public void run() {

		if ( nick.equals("PROMPT")) 
			waitfornick();

		if ( !connect())
			return;

		System.out.println("Waiting for data...\n");
		tb.wait_notice("Waiting for data...");
		boolean waiting=true;

		String str;
		while ( (runner != null) && ((str = getline() ) != null) ) {
			if ( waiting ) {
				waiting = false;
				tb.notice();
			}
			System.out.println( str );
			parse(str);
		}

		System.out.println("Closing...\n");
		close();
	}

	/*
		This should be encapsulated in another class of course
	*/
	private void parse(String str) {
		String head=null, body=null, nick=null;
		int sep;

		// Trim leading :
		if ( str.startsWith(":") )
			str = str.substring(1, str.length());

		if ( (sep = str.indexOf((int)':')) != -1 ) {
			head = str.substring(0, sep);
			if ( str.length() > sep+1 )
				body = str.substring(sep+1, str.length());
		} else
			head = str;


		if ( (sep = head.indexOf((int)'!')) !=- 1  )
			nick = head.substring(0, sep);

		System.out.println("Head = " + head);
		System.out.println("Body = " + body);
		System.out.println("Nick = " + nick);
	
		if ( head.indexOf("PING") != -1 ) {
			System.out.println("Got PING request\n");
			if ( keepalive && !STOPPED )
				System.out.println("Sent PONG");
				out.println("pong "+body);
			return;
		}

		if ( head.indexOf("JOIN") != -1 ) {
			// Body has channel, but doesn't on PART
			tb.put(Color.red, "<"+nick+"> has joined the channel");
			return;
		}
		if ( head.indexOf("PART") != -1 ) {
			tb.put(Color.red, "<"+nick+"> has left the channel");
			return;
		}
		if ( head.indexOf("QUIT") != -1 ) {
			tb.put(Color.red, "<"+nick+"> has quit");
			return;
		}
		if ( head.indexOf("NICK") != -1 ) {
			tb.put(Color.red, 
				"<"+nick+"> has changed his nickname to <"+body+">");
			return;
		}
		if ( head.indexOf("MODE") != -1 ) {
			if (nick != null)
				tb.put(Color.red, "<"+nick+"> change mode: ["+head+"]");
			return;
		}
		if ( head.indexOf("KICK") != -1 ) {
			tb.put(Color.red, "<"+nick+"> "+body);
			tb.put(Color.red, 
				"*** <"+nick+"> has kicked you off the channel ***");
			return;
		}
		if ( head.indexOf("PRIVMSG") != -1 ) {
			tb.put("<"+nick+"> "+body);
			return;
		}
		if ( head.indexOf("ERROR") != -1 ) {
			if ( body.indexOf("Closing") != -1 )
				tb.put(Color.red, "*** Connection Closed ***");
			return;
		}

		if (body!=null)
			if (nick!=null)
				tb.put("<"+nick+"> "+body);
			else
				tb.put("<?> "+body);
		else
			if (head!=null)
				tb.put("["+head+"] ");
			else
				tb.put("[?]");
	}

	private String getline() {
	    StringBuffer buff = new StringBuffer();
		int ch;

		try {
			while ( (ch = in.read()) != -1 ) {
				if (ch == '\n')
					break;
				else
					buff.appendChar(ch);
			}
		} catch (Exception e) {
			System.out.println("in.read exception: " + e);
		}

		if ( buff.length() > 0 )
			return(buff.toString());
		else
			return ( null );
	}

	public void stop() {
		STOPPED=true;
	}

    void close() {
		try {
			nc.closeServer();
		} catch (Exception e) {
			System.out.println("close server exception: "+e);
		}

		nc = null;
    }

	// (almost) random 9 char nickname
	private String randnick() {
		try {
			Math.srandom((int)System.currentTimeMillis());
		} catch (Exception e) {
			System.out.println("Trying nowMillis()");
			try {
				Math.srandom((int)System.nowMillis());
			} catch (Exception e2) {
				System.out.println("Cant get Millis... winging it\n");
			}
		}

		StringBuffer s = new StringBuffer("Foo");
		for (int i=0; i<6; i++) {
			char ch = (char)(Math.random()*26 + 65);
			s.append(new Character(ch).toString());
		}
		return(s.toString());
	}

	public synchronized void EntryLine(String s) {

		System.out.println("In EntryLine, s = "+s);
		if ( (s == null) || ( s.length() == 0 ) )
			return;

		if ( nick.equals("PROMPT") ) {
			String n = s.substring(te.prompt);
			System.out.println("nick = PROMPT, got entry = "+n);
			if ( n.length() > 0 ) {
				nick = n;
				notify();
				System.out.println("Did notify from Entryline");
			}
			return;
		}

		if ( nc == null )
			return;

		if ( s.startsWith("/") ) {
			try {
				out.println(s.substring(1));
				tb.put(Color.red, s.substring(1));
			} catch (Exception e ) {
			}
			return;
		}
		tb.put(Color.white, "> "+s);
		try {
			out.println("privmsg "+channel+" :"+s);
		} catch (Exception e) {
		}
	}

	/* 
		For our simple text classes we just pass relevant events explicitly.
	*/
	public void update(Graphics g) {
		tb.update(g);
		te.update(g);
	}
	public void paint(Graphics g) {
		tb.paint(g);
		te.paint(g);
	}
	public void mouseDown(int x, int y) {
		tb.mouseDown(x,y);
		te.mouseDown(x,y);
	}
	public void mouseUp(int x, int y) {
		tb.mouseUp(x,y);
	}
	public void mouseDrag(int x, int y) {
		tb.mouseDrag(x,y);
	}
	public void keyDown(int ch) {
		te.keyDown(ch);
	}

}
